/* /usr/local/plot79/prim/xio.c, Tue Jan  1 01:04:18 1991 */
/* Edit by Nelson H. F. Beebe <beebe@ibm.math.utah.edu> */
/***********************************************************************
Berkeley-UNIX style Fortran character I/O for IBM AIX and other UNIX
systems that lack the possibility of supporting the Berkeley routines
directly (e.g. any UNIX Fortran implementation that does not produce
different external names than the C compiler does for the same name).

The unit number passed as an argument to these routines is a small
non-negative integer in the range 0..99.  Three units are
preconnected: 0 (stderr), 5 (stdin), and 6 (stdout).  They may be
reused for other files if they are first closed.  The unit number has
NO connection whatever with Fortran unit numbers, and I/O to the unit
numbers passed to these routine cannot be done with Fortran I/O
statements.

The function return value RET_XEOF is guaranteed to be negative.

void xclear(FINT u)
	Clear the error indicator on unit u.

int xclose(FINT u)
	Close the file open on unit u, and return u on success, or
	RET_XEOF on failure.  If the unit is already closed, return u.

int xeof(FINT u)
	Return non-zero if and only if the end-of-file has been
	reached on unit u, and zero otherwise (including illegal u).

int xerr(FINT u)
	Return non-zero if and only if the error indicator for unit
	u, and zero otherwise (including illegal u).

int xfilno(FINT u)
	Return the UNIX integer file descriptor for unit u, or RET_XEOF
	on error.  [This can be used in various UNIX system calls.]

int xflush(FINT u)
	Flush the output buffer for unit u, and return u on success,
	or RET_XEOF on failure.

int xfname(FINT u, FCHAR filename)
	Return in filename the name of the file currently open on
	unit u (the one passed to xopen()), and return u on success,
	or RET_XEOF on failure.

int xgetc(FINT u)
	Return the next input character from unit u, or RET_XEOF at
	end-of-file or error.

int xistty(FINT u)
	Return zero if u is open on a terminal device, and RET_XEOF otherwise.

int xopen(FINT u, FCHAR filename, FCHAR mode)
	Open unit u for filename with fopen mode, and return u, or
	RET_XEOF on error.

int xputb(FINT u, FINT c)
	Output byte c on unit u, and return c, or RET_XEOF on error.

int xputc(FINT u, FCHAR c)
	Output character c on unit u, and return c, or RET_XEOF on error.

int xputh(FINT u, char *h, FINT nh)
	Output nh characters of the Hollerith string h on unit u, and
	return u, or RET_XEOF on error.

int xputs(FINT u, FCHAR s)
	Output CHARACTER string s on unit u, and return u,
	or RET_XEOF on error.  The string is NOT NULL terminated; its
	length is passed by the Fortran code.

int xunit()
	Return a free unit number for use in a call to xopen(),
	or RET_XEOF on failure.

Note that the implementation of the routines with char* arguments must
know how Fortran strings are passed to C; the current implementation
assumes that the lengths are passed as additional arguments.  For
xputc(), the length modifier is ignored.  Modifications will be needed
if character string passing is by descriptor.

In this implementation, conversion between C and Fortran strings
is done by temporary writes of a NUL terminator; if the Fortran
compiler puts constant strings in read-only memory, then minor
code changes will be needed here.

[23-Apr-1991] -- add AIX370 support (different numbers of args if any
		 arg is of type CHARACTER)
[26-Mar-1991] -- bug fix for Stardent
[15-Feb-1991]
***********************************************************************/

#include <stdio.h>
#include <string.h>

#if  __STDC__
#include <stdlib.h>
#endif

#if _AIX370
#include <malloc.h>
int isatty (int fildes);
#endif

/* CONS(a,b) should return ab, the concatenation of its arguments */
#if  __STDC__ || __APOGEE__
#define CONS(a,b) a##b
#else /* NOT __STDC__ || __APOGEE__ */
#define CONS(a,b) a/**/b
#endif /* __STDC__ || __APOGEE__ */

/* To simplify modifications for other systems, we use typedefs
for Fortran INTEGER (FINT) and CHARACTER (FCHAR) data, and
access macros CH(),  REF(), and VAL().   The access macros
must be usable as both l-values and r-values.  */

#if ardent
/* Stardent Fortran passes CHARACTER data by descriptor. */
typedef int*	FINT;
typedef struct
{
    char        *address;
    int         length;
} *FCHAR;
#define CH(x,n)	((x->address)[n])/* n-th character of x */
#define FLEN(x)	(x->length)	/* declared length of CHARACTER string */
#define FLENARG(x) CONS(x,_len)	/* declared length of CHARACTER string */
#endif /* ardent */

#if _AIX370
typedef int*	FINT;
typedef char*	FCHAR;
#define CH(x,n)	x[n]		/* n-th character of x */
#define FLEN(x)	*(CONS(x,_len))	/* declared length of CHARACTER string */
#define FLENARG(x) CONS(x,_len)	/* declared length of CHARACTER string */
#endif

#if __hp9000s500
typedef int*	FINT;
typedef struct
{
    int		max_length;
    char	*curlenword;
    int		first_byte;
    int         length;
    char        *address;
} *FCHAR;
#define CH(x,n)	((x->address)[n])/* n-th character of x */
#define FLEN(x)	(x->length)	/* declared length of CHARACTER string */
#define FLENARG(x) CONS(x,_len)	/* declared length of CHARACTER string */
#endif /* __hp9000s500 */

#ifndef FLENARG
/* Most UNIX compilers pass CHARACTER data by the address of the first
byte, with the length passed as an additional argument at the end of the
Fortran-visible argument list.  The HP 9000/800 passes the lengths
by value immediately following each CHARACTER argument. */
typedef int*	FINT;
typedef char*	FCHAR;
#define CH(x,n)	x[n]		/* n-th character of x */
#define FLEN(x)	CONS(x,_len)	/* declared length of CHARACTER string */
#define FLENARG(x) CONS(x,_len)	/* declared length of CHARACTER string */
#endif

#define REF(x)	x		/* reference to (address of) x */
#define VAL(x)	*x		/* value of x */

#define RET_XEOF (-1)		/* return value on end-of-file or error */

/* Units 0, 5, and 6 are preconnected, and need not be explicitly opened */

static FILE* fp_table[] =
{
	 stderr, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
	  stdin,      stdout, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
    (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL, (FILE*)NULL,
};
static char* filename_table[] =
{
    "/dev/stderr", (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    "/dev/stdin",  "/dev/stdout", (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
    (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,   (char*)NULL,
};

#define FP(u)			fp_table[VAL(u)]

#define IS_OPEN(u)		(FP(u) != (FILE*)NULL)

#define LEGAL(u)		((0 <= u) && (u < MAXUNIT))

#define LEGAL_AND_OPEN_FP(u)	(LEGAL(VAL(u)) && IS_OPEN(u))

#define MAXUNIT			(sizeof(fp_table)/sizeof(fp_table[0]))

#define MIN(a,b)		( ((a) < (b)) ? (a) : (b) )


#ifdef ardent
#define	xclear	XCLEAR
#define	xclose	XCLOSE
#define	xeof	XEOF
#define	xerr	XERR
#define	xfilno	XFILNO
#define	xflush	XFLUSH
#define	xfname	XFNAME
#define	xgetc	XGETC
#define	xistty	XISTTY
#define	xopen	XOPEN
#define	xputb	XPUTB
#define	xputc	XPUTC
#define	xputh	XPUTH
#define	xputs	XPUTS
#define	xunit	XUNIT
#endif

#if _AIX370 || __mips || __sgi
#define	xclear	xclear_
#define	xclose	xclose_
#define	xeof	xeof_
#define	xerr	xerr_
#define	xfilno	xfilno_
#define	xflush	xflush_
#define	xfname	xfname_
#define	xgetc	xgetc_
#define	xistty	xistty_
#define	xopen	xopen_
#define	xputb	xputb_
#define	xputc	xputc_
#define	xputh	xputh_
#define	xputs	xputs_
#define	xunit	xunit_
#endif

static void
FtoC(f,nf)			/* Convert Fortran string to C string */
char *f;			/* by changing first trailing blank to NUL */
int nf;
{
    char *p;

    p = &f[nf-1];
    while ((*p == ' ') && (p > f))
	--p;
    p[1] = '\0';		/* provide NULL terminator */
}

static void
CtoF(c)				/* convert C string to Fortran string */
char *c;			/* where the Fortran string already */
{				/* has all needed trailing blanks */
    char *p;

    p = strchr(c,'\0');
    *p = ' ';
}

void
xclear(u)
FINT u;				/* unit number */
{
    if (LEGAL_AND_OPEN_FP(u))
	clearerr(FP(u));
}


int
xclose(u)			/* return RET_XEOF on failure, u on success */
FINT u;				/* unit number */
{
    int code;

    if (LEGAL(VAL(u)))
    {
	if (IS_OPEN(u))
	{
	    code = fclose(FP(u));	/* NB: no garbage collection */
	    filename_table[VAL(u)] = (char*)NULL; /* on malloc'ed storage */
	    FP(u) = (FILE*)NULL;
	}
	else
	    code = 0;		/* already closed */
    }
    else			/* illegal unit number */
	code = RET_XEOF;
    return ((code == 0) ? VAL(u) : RET_XEOF);
}

int
xeof(u)
FINT u;				/* unit number */
{
    if (LEGAL_AND_OPEN_FP(u))
	return (feof(FP(u)));
    else
	return (0);
}

int
xerr(u)
FINT u;				/* unit number */
{
    if (LEGAL_AND_OPEN_FP(u))
	return (ferror(FP(u)));
    else
	return (0);
}

int
xfilno(u)
FINT u;				/* unit number */
{
    if (LEGAL_AND_OPEN_FP(u))
	return (fileno(FP(u)));
    else
	return (-1);
}

int
xflush(u)
FINT u;
{
    if (LEGAL_AND_OPEN_FP(u))
	return ((fflush(FP(u)) < 0) ? RET_XEOF : VAL(u));
    else
	return (RET_XEOF);		/* illegal unit number */
}

#if _AIX370
int
xfname(u, filename, FLENARG(u), FLENARG(filename))
FINT u;
FCHAR filename;
FINT FLENARG(u);
FINT FLENARG(filename);
#else
int
xfname(u, filename, FLENARG(filename))
FINT u;
FCHAR filename;
int FLENARG(filename);
#endif
{
    int n;
    (void)memset(&CH(filename,0), ' ', FLEN(filename));
    if (LEGAL_AND_OPEN_FP(u))
    {
	n = strlen(filename_table[VAL(u)]);
	(void)strncpy(&CH(filename,0), filename_table[VAL(u)], 
	    MIN(FLEN(filename),n));
	CH(filename,FLEN(filename)-1) = '\0';/* in case strncpy did not do it */
	CtoF(&CH(filename,0));
	CH(filename,FLEN(filename)-1) = ' ';
	return (VAL(u));
    }
    else
	return (RET_XEOF);		/* illegal unit number */
}

int
xgetc(u,c)		/* return RET_XEOF on failure or EOF, c on success */
FINT u;				/* unit number */
FCHAR c;
{
    int ch;

    if (LEGAL_AND_OPEN_FP(u))
	ch = getc(FP(u));
    else
	ch = RET_XEOF;
    CH(c,0) = ch;
    return (ch);
}

int
xistty(u)
FINT u;
{
    return ( (LEGAL(VAL(u)) && (isatty(fileno(FP(u))) == 1)) ? 0 : RET_XEOF );
}

#if _AIX370
int				/* return RET_XEOF on failure, u on success */
xopen(u, filename, mode, FLENARG(u), FLENARG(filename), FLENARG(mode))
FINT u;				/* unit number */
FCHAR filename;			/* Fortran file name */
FCHAR mode;			/* open mode (any valid C mode for fopen) */
FINT FLENARG(u);		/* len(u) */
FINT FLENARG(filename);		/* len(filename) */
FINT FLENARG(mode);		/* len(mode) */
#else
#if __hp9000s800
int				/* return RET_XEOF on failure, u on success */
xopen(u, filename, FLENARG(filename), mode, FLENARG(mode))
#else
int				/* return RET_XEOF on failure, u on success */
xopen(u,filename,mode, FLENARG(filename),FLENARG(mode))
#endif
FINT u;				/* unit number */
FCHAR filename;			/* Fortran file name */
FCHAR mode;			/* open mode (any valid C mode for fopen) */
int FLENARG(filename);		/* len(filename) */
int FLENARG(mode);		/* len(mode) */
#endif
{
    if (LEGAL(VAL(u)))
    {
	FtoC(&CH(filename,0),FLEN(filename));
	FtoC(&CH(mode,0),FLEN(mode));
	FP(u) = fopen(&CH(filename,0),&CH(mode,0));
	if (IS_OPEN(u))
	{			/* save name for xfname() */
	    filename_table[VAL(u)] = (char*)malloc(strlen(&CH(filename,0))+1);
	    if (filename_table[VAL(u)] != (char*)NULL)
		(void)strcpy(filename_table[VAL(u)],&CH(filename,0));
	}
	CtoF(&CH(filename,0));
	CtoF(&CH(mode,0));

	return (IS_OPEN(u) ? VAL(u) : RET_XEOF);
    }
    else
	return (RET_XEOF);
}

int
xputb(u,c)			/* return RET_XEOF on failure, c on success */
FINT u;				/* unit number */
FINT c;				/* character */
{
    if (LEGAL_AND_OPEN_FP(u))
    {
	putc(VAL(c),FP(u));
	return (VAL(u));
    }
    else
	return (RET_XEOF);
}

int
xputc(u,c)			/* return RET_XEOF on failure, c on success */
FINT u;				/* unit number */
FCHAR c;			/* character */
{
    if (LEGAL_AND_OPEN_FP(u))
    {
	putc(CH(c,0),FP(u));
	return (VAL(u));
    }
    else
	return (RET_XEOF);
}

int
xputh(u,h,nh)			/* return RET_XEOF on failure, u on success */
FINT u;				/* unit number */
char *h;			/* Hollerith string */
FINT nh;			/* length of h[] */
{
    int code;
    char c;

    if (LEGAL_AND_OPEN_FP(u))
    {
	c = h[VAL(nh)];
	h[VAL(nh)] = '\0';
	code = fputs(h,FP(u));
	h[VAL(nh)] = c;
	return ((code < 0) ? RET_XEOF : VAL(u));
    }
    else
	return (RET_XEOF);
}

#if _AIX370
int
xputs(u,s,FLENARG(u),FLENARG(s))/* return RET_XEOF on failure, u on success */
FINT u;				/* unit number */
FCHAR s;			/* CHARACTER string */
FINT FLENARG(u);		/* (hidden) length of u */
FINT FLENARG(s);		/* (hidden) length of s[] */
#else
int
xputs(u,s,FLENARG(s))		/* return RET_XEOF on failure, u on success */
FINT u;				/* unit number */
FCHAR s;			/* CHARACTER string */
int FLENARG(s);				/* (hidden) length of s[] */
#endif
{
    int code;
    char c;

    if (LEGAL_AND_OPEN_FP(u))
    {
	c = CH(s,FLEN(s));
	CH(s,FLEN(s)) = '\0';
	code = fputs(&CH(s,0),FP(u));
	CH(s,FLEN(s)) = c;
	return ((code < 0) ? RET_XEOF : VAL(u));
    }
    else
	return (RET_XEOF);
}

int
xunit()			/* return free unit number, or RET_XEOF on failure  */
{
    int u;

    for (u = (MAXUNIT - 1); u >= 0; --u)
    {
	if (!IS_OPEN(&u))
	    return (u);
    }
    return (RET_XEOF);    
}
