#!/bin/csh  -f
# Unix csh version of sf3 preprocessor control
#
# The global variables SF3FLAGS  and  FFLAGS  may  be  used  to  provide
# switches  for sf3 and f77.   In addition, if the file "sf3.opt" exists
# in the current or HOME directory, its contents will be appended to the
# command  line.   Since  option  switches are scanned for from  left to
# right, the SF3FLAGS values will override any provided in the "sf3.opt"
# file.  The default Fortran compiler  can be overridden  by setting the
# environment variable FC.
#
# Usage: sf3 [-s1 -s2 ... -sn] filelist
#
# Example: sf3 -w:72 -k -f a*.sf3 *bar*.sf3
#
# The available switches are:
#
# -C	- produce CDC-style code for READ/WRITE END and ERR parameters
# -D	- debug - include sf3 code as comments in output Fortran code
# -F	- invoke the Fortran compiler on the output ".f" files; compiler
#	  options may be supplied in the FFLAGS global variable
# -H	- display a help file and terminate immediately
# -I:n	- set structure indentation to n (0..9).
# -K	- keep ".f" files after compilation
# -L	- set the LINEUP option which connects structure begin/end with
#	  dotted lines in the ".sfl" files
# -M:nn - set maximum lines/page to nn (suggest 58 for 6 lpi and 80 for
#         8 lpi); ignored when prettyprinting.
# -N	- no ".sfl" files
# -P	- prettyprinted source code sent to ".sfl" files instead of a
#	  normal listing
# -R	- reprocess everything, even if ".f" or ".o" files are newer
#	  than ".sf3" files
# -W:nn	- set ".sfl" file line width to nn (72..120)
#
# If  the  "-k"  option  is  specified  on  the command line,  or in the
# SF3FLAGS variable, or in the "sf3.opt" file, the Fortran files will be
# kept.    By  default, they are deleted after a successful compilation.
# If the compilation fails, however, they are preserved.
#
# On completion, the ".sf3" files will  be  converted  to  ".f"  Fortran
# files,  with listings  in ".sfl" files, and if the compile option "-f"
# was given, ".o" object files will have been  made.   File  dates/times
# are compared to determine whether to preprocess any ".sf3" file or run
# the Fortran compiler, unless this action  is  overriden  by  the  "-r"
# option.
#
# If any filename has no extension, ".sf3" will be assumed.
#
# [11-Jun-1990] - allow Fortran compiler to be set from environment
# [05-Sep-1984]
# [16-Feb-1985] - minor mods for 4.2BSD at UCSF
# [31-Jul-1987] - changed sf3 to sf3.exe (avoids name conflict in sf3dir)
#=======================================================================
# Berkeley 4.1BSD (Eunice under VAX VMS)
# An apparent bug in CSH causes it to incorrectly match an inner  "else"
# with  a  failing  outer  "then"  if  these  keywords  are followed  by
# comments.  Moving the comments to separate lines fixed the problem.
#
# Another bug in f77 causes it to be unable to  handle  /dev/null  which
# sf3  assumes  if a file is missing on the command line.   We therefore
# create a temporary empty file for this purpose.
# [05-Sep-1984]
#=======================================================================

# We have sf3.exe, sf3.opt, and sf3.hlp in this directory
set SF3DIR=/usr/local/plot79
set TEMPFILE=/tmp/sf3$$
if ($?SF3FLAGS == 0) set SF3FLAGS=
if ($?FFLAGS == 0) set FFLAGS=
if ($?FC == 0) then
    foreach COMPILER__ ( /usr/lang/f77 /bin/f77 /usr/bin/f77 \
			/usr/local/f77 /bin/fc /usr/bin/fvs /usr/bin/xlf )
	if (-e $COMPILER__) then
	    set FC=$COMPILER__
	    break
	endif
    end
endif
if ($?FC == 0) then
    echo '?Cannot find Fortran compiler'
    exit 1
endif

# Append any flags from the "sf3.opt" file to SF3FLAGS string

set CMDFLAGS="$SF3FLAGS"
if (-e sf3.opt) then
    set CMDFLAGS="$CMDFLAGS `cat sf3.opt`"
else if (-e $SF3DIR/sf3.opt) then
    set CMDFLAGS="$CMDFLAGS `cat $SF3DIR/sf3.opt`"
endif

# Now separate the switches from the file names

set FILELIST=
foreach FULLNAME ($*)
    switch ($FULLNAME)
    case "-*":
        set CMDFLAGS="$CMDFLAGS $FULLNAME"
        breaksw
    default:
        set FILELIST="$FILELIST $FULLNAME"
        breaksw
    endsw
end

# Append the default switches to the explicit ones and note special ones
# that we handle here at shell level instead of inside the preprocessor

set COMPILE=0 KEEPF=0 KEEPSFL=1 PRETTY=0 REDO=0 SF3STATUS=0
foreach FULLNAME ($CMDFLAGS)
    switch ($FULLNAME)
    case -[fF]:
        set COMPILE=1
        breaksw
    case -[hH]:
        more $SF3DIR/sf3.hlp
        exit(0)
    case -[kK]:
        set KEEPF=1
        breaksw
    case -[nN]:
        set KEEPSFL=0
        breaksw
    case -[pP]:
        set PRETTY=1
        breaksw
    case -[rR]:
        set REDO=1
    default:
        breaksw
    endsw
end
if ($PRETTY == 1) set KEEPSFL=1

# main loop over ".sf3" file names - process them one at a time

foreach FULLNAME ($FILELIST)
    if ("$FULLNAME:e" == "") set FULLNAME=${FULLNAME}.sf3
    if ($FULLNAME:h == $FULLNAME:t) then
        set PATHNAME=			# no directory path
    else
        set PATHNAME="$FULLNAME:h"/
    endif
    if (-e $FULLNAME) then
        set ROOTNAME=$FULLNAME:r
        echo \[$ROOTNAME\]
        if ($REDO == 1) then
            set DOF77=1 DOSF3=1
        else
            set DOF77=0 DOSF3=0
            if (-e ${ROOTNAME}.o) then
                set ORDER=(`ls -t ${ROOTNAME}.sf3 ${ROOTNAME}.o`)
                if ("$ORDER[1]" == "${ROOTNAME}.sf3") then
                    set DOSF3=1 DOF77=1 		# .sf3 is newer
                else
                    echo "   ${ROOTNAME}.o is up-to-date" # .o is newer
                endif
            else
    							# no .o file
                if (-e ${ROOTNAME}.f) then
    							# have .f file already
                    set ORDER=(`ls -t ${ROOTNAME}.sf3 ${ROOTNAME}.f`)
                    if ("$ORDER[1]" == "${ROOTNAME}.sf3") then
                        set DOSF3=1			# .sf3 newer
                    else
                        echo "   ${ROOTNAME}.f is up-to-date" # .f is newer
                    endif
                else
                    set DOSF3=1 # no .f file
                endif
                set DOF77=1
            endif
        endif
        if ($DOSF3 == 1) then
            if (-e ${ROOTNAME}.inc) then
                set INCLUDEFILE=${ROOTNAME}.inc
            else if (-e ${PATHNAME}sftran.inc) then
                set INCLUDEFILE=${PATHNAME}sftran.inc
            else if (-e ${PATHNAME}sf3.inc) then
                set INCLUDEFILE=${PATHNAME}sf3.inc
            else
		cat /dev/null >$TEMPFILE
                set INCLUDEFILE=$TEMPFILE
            endif

            $SF3DIR/sf3.exe <$FULLNAME >${ROOTNAME}.f "${ROOTNAME}.sfl" $CMDFLAGS "$INCLUDEFILE"
            # suppress compilation if unsuccessful, and remove any
            # empty or unrequired files
            if ($status > $SF3STATUS) set SF3STATUS=$status
            if ($status != 0) set DOF77=0
            if (-z ${ROOTNAME}.f) rm -f ${ROOTNAME}.f
            if ((KEEPSFL == 0) || (-z ${ROOTNAME}.sfl)) rm -f ${ROOTNAME}.sfl
        endif
        if (($COMPILE == 1) && ($DOF77 == 1) && (-e ${ROOTNAME}.f)) then
	    echo $FC $FFLAGS -c ${ROOTNAME}.f
            $FC $FFLAGS -c ${ROOTNAME}.f
            if ($status > $SF3STATUS) set SF3STATUS=$status
            if (($KEEPF == 0) && ($SF3STATUS ==  0)) rm -f ${ROOTNAME}.f
        endif
    else
        echo "? $FULLNAME does not exist"
    endif
end # of foreach loop
rm -f $TEMPFILE
exit($SF3STATUS)
